-- Monads.
-- Author :: Ulf Norell
-- Created:: 2004-12-14

-- # import Property
-- # import Prelude
-- # import List

-- The definition of a monad ----------------------------------------------

MonadLaws :  (M : Set -> Set)                        ->
            ((A:Set) -> A -> M A)                    ->
            ((A,B:Set) -> M A -> (A -> M B) -> M B)  ->
            Type
MonadLaws M return bind = sig
  law1 : (A:Set) -> (m : M A) ->

            let (>>=) = bind A A
                ret   = return A
            in
                (m >>= ret) == m

  law2 : (A,B:Set) -> (x : A) -> (f : A -> M B) ->

            let (>>=) = bind A B
                ret   = return A
            in
                (ret x >>= f) == f x

  law3 : (A,B,C:Set) -> (m : M A) -> (g : A -> M B) -> (h : B -> M C) ->

            let (>>=)   = bind A B
                (>>=')  = bind B C
                (>>='') = bind A C
            in
                ((m >>= g) >>=' h) == (m >>='' (\x -> g x >>=' h))
  

MonadOps : (M : Set -> Set) -> Type
MonadOps M = sig
  return : (A : Set) -> A -> M A
  bind   : (A, B : Set) -> M A -> (A -> M B) -> M B
  

Monad : (M : Set -> Set) -> Type
Monad M = sig
  ops    : MonadOps M
  proofs : MonadLaws M ops.return ops.bind
  
-- Monad functions --------------------------------------------------------

return : (M : Set -> Set) -> MonadOps M -> (A : Set) -> A -> M A
return M monadM = monadM.return

bind : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
        M A -> (A -> M B) -> M B
bind M monadM = monadM.bind

bind_ : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
          M A -> M B -> M B
bind_ M monadM A B m m' = monadM.bind A B m (\ _ -> m')

ap : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
      M (A -> B) -> M A -> M B
ap M monadM A B mf ma =
  let (>>=)   = monadM.bind (A -> B) B
      (>>=')  = monadM.bind A B
      return  = monadM.return B
  in  mf >>=  \f ->
      ma >>=' \a ->
      return (f a)

join : (M : Set -> Set) -> MonadOps M -> (A : Set) ->
        M (M A) -> M A
join M monadM A m =
  let (>>=)   = monadM.bind (M A) A
  in  m >>= id

liftM : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
          (A -> B) -> M A -> M B
liftM M monadM A B f m =
  let (>>=)   = monadM.bind A B
      return  = monadM.return B
  in  m >>= \x -> return (f x)

liftM2 : (M : Set -> Set) -> MonadOps M -> (A, B, C : Set) ->
          (A -> B -> C) -> M A -> M B -> M C
liftM2 M monadM A B C f ma mb =
  let (<$>)   = liftM M monadM A (B -> C)
      (<*>)   = ap M monadM B C
  in  f <$> ma <*> mb

liftM3 : (M : Set -> Set) -> MonadOps M -> (A, B, C, D : Set) ->
          (A -> B -> C -> D) -> M A -> M B -> M C -> M D
liftM3 M monadM A B C D f ma mb mc =
  let (<$>)   = liftM M monadM A (B -> C -> D)
      (<*>)   = ap M monadM B (C -> D)
      (<*>')  = ap M monadM C D
  in  f <$> ma <*> mb <*>' mc

sequence : (M : Set -> Set) -> MonadOps M -> (A : Set) ->
            List (M A) -> M (List A)
sequence M monadM A =
  let liftM2  = liftM2 M monadM A (List A) (List A)
      return  = monadM.return (List A)
  in  foldr (liftM2 (::)) (return [])

sequence_ : (M : Set -> Set) -> MonadOps M -> (A : Set) ->
            List (M A) -> M Unit
sequence_ M monadM A =
  let (>>)    = bind_ M monadM A Unit
      return  = monadM.return Unit
  in  foldr (>>) (return unit)

mapM : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
        (A -> M B) -> List A -> M (List B)
mapM M monadM A B f xs =
  let sequence  = sequence M monadM B
  in  sequence (map f xs)

mapM_ : (M : Set -> Set) -> MonadOps M -> (A, B : Set) ->
        (A -> M B) -> List A -> M Unit
mapM_ M monadM A B f xs =
  let sequence_ = sequence_ M monadM B
  in  sequence_ (map f xs)

