
\section{Proving things about QuickCheck generators} \label{QuickCheck/Proofs.lagda}

\begin{code}
-- # import QuickCheck
-- # import List.Proofs
-- # import Bool.Proofs
-- # import Nat.Proofs
-- # import Ord.Proofs
\end{code}

Combining QuickCheck with a system such as AgdaLight means, not only that we
can test properties before proving them, but also that we can prove properties
about the tests. Something, in particular, that we might be interested in is
generator surjectivity, i.e. we want to know that a generator can generate all
possible values satisfying a particular property.

%- Axioms -----------------------------------------------------------------
\subsection{Random number axioms}

We postulate surjectivity of \<split\>. This is not {\em really} true, since
there aren't an infinite number of random seeds (although you could certainly
imagine an implementation of random numbers for which this is true). For
practical purposes, though, there are enough of them.

\begin{code}
propSplitSurjective :
    (r1, r2 : StdGen) ->
    CanProduce {StdGen;StdGen} split (r1,r2)
\end{code}

Pick can produce any element in the list.

\begin{code}
propPickSurjective :
    (A:Set) -> (xs:List A) -> (x:A)  ->
    PropElem A x xs                  ->
    CanProduce A (pick A xs) x
\end{code}

%- Proofs about random numbers --------------------------------------------
\subsection{Proofs about the $Rand$ monad}

\begin{code}
propReturnCanProduce :
    (A : Set) -> (x : A) -> CanProduce A (monadR.return A x) x
propReturnCanProduce A x = (seed, reflexivity A x)

propBindCanProduce :
    (A, B : Set) -> (g : Rand A) -> (k : A -> Rand B) ->
    (x : A) -> (y : B) ->
    CanProduce A g x -> CanProduce B (k x) y ->
    CanProduce B (monadR.bind A B g k) y
propBindCanProduce A B g k x y pg pk =
    (splitr.1, fol-plugin(pg.2, pk.2, splitr.2, __fol_pair_elimination))
  where
    splitr  = propSplitSurjective pg.1 pk.1
\end{code}

%- Proofs about the standard generators -----------------------------------
\subsection{Proofs about the standard generators}

\<bind\> is surjective.. no it's not!
\<g\> and \<k\> might only be able to generate \<x\> and \<y\>
for very particular (and disjoint) sizes.

\begin{code}
propBindCanGenerate_Wrong :
    (A, B : Set) -> (g : Gen A) -> (k : A -> Gen B) ->
    (x : A) -> (y : B) ->
    CanGenerate A g x -> CanGenerate B (k x) y ->
    CanGenerate B (monadGen.bind A B g k) y
propBindCanGenerate_Wrong A B g k x y pg pk = assert-plugin(silent)
\end{code}

The property (I think) we want is::

\begin{code}
CanGenerate' : (A : Set) -> (g : Gen A) -> (x : A) -> Type
CanGenerate' A g x =
  { size : Nat
  ; (size' : Nat) -> Holds (ordNat.leq size size') ->
                      CanProduce A (g size') x
  }
\end{code}

  This should be more compositional, although it might be a lot more work
  establishing it.

  The question is what it would mean for the proofs? For instance this will not
  hold (for any result) of the generator \<sized return\>, but that might not
  be a big problem since that generator is fundamentally flawed and it doesn't
  matter if you can't prove anything about it(?)

  Anyway we can easily prove that it entails the original property.
\begin{code}

propCanGenerate' :
  (A : Set) -> (g : Gen A) -> (x : A) ->
  CanGenerate' A g x -> CanGenerate A g x
propCanGenerate' A g x p = (p.1, p.2 p.1 (reflOrd Nat ordNat p.1))

\end{code}

  It is a trivial proof that \<return\> preserves \<CanGenerate'\>.

\begin{code}

-- return preserves CanGenerate'
propReturnCanGenerate :
    (A : Set) -> (x : A) -> CanGenerate' A (monadGen.return A x) x
propReturnCanGenerate A x = (0, \_ _ -> propReturnCanProduce A x)

\end{code}

  The proof that \<bind\> preserves \<CanGenerate'\> isn't that bad actually.
  Given the nice theorems we have about \<\leq\> and \<max\> it goes through
  quite smoothly.

\begin{code}

-- bind preserves CanGenerate'
propBindCanGenerate :
    (A, B : Set) -> (g : Gen A) -> (k : A -> Gen B) ->
    (x : A) -> (y : B) ->
    CanGenerate' A g x -> CanGenerate' B (k x) y ->
    CanGenerate' B (monadGen.bind A B g k) y
propBindCanGenerate A B g k x y pg pk =
    ( size
    , \size' leq ->
        propBindCanProduce A B (g size') (\z -> k z size') x y
          (pg.2 size' fol-plugin( leq
                                , propMaxLeq Nat ordNat
                                , ordNat.proofs.trans
                                )
          )
          (pk.2 size' fol-plugin( leq
                                , propMaxLeq Nat ordNat
                                , ordNat.proofs.trans
                                , propMaxCommute Nat ordNat
                                , propNatEq
                                )
          )
    )
  where
    size = max Nat ordNat pg.1 pk.1

propLiftMCanGenerate :
    (A, B : Set) -> (f : A -> B) -> (g : Gen A) -> (x : A) ->
    CanGenerate' A g x -> CanGenerate' B (liftM Gen monadGen A B f g) (f x)
propLiftMCanGenerate A B f g x p =
    propBindCanGenerate A B
        g (\y -> monadGen.return B (f y))
        x (f x) p (propReturnCanGenerate B (f x))

\end{code}

\begin{code}

-- rand is surjective
propRandCanGenerate : (r : StdGen) -> CanGenerate' StdGen rand r
propRandCanGenerate r = (0, \_ _ -> (r, reflexivity StdGen r))

-- elements is surjective
propElementsCanGenerate :
    (A : Set) -> (xs : List A) -> (x : A) ->
    PropElem A x xs ->
    CanGenerate' A (elements A xs) x
propElementsCanGenerate A xs x el =
    propBindCanGenerate StdGen A
      rand (\r -> monadGen.return A (pick A xs r))
      p.1  x
      (propRandCanGenerate p.1)
      (0, \_ _ -> (seed, p.2))
  where
    p = propPickSurjective A xs x el

-- oneof is surjective
propOneofCanGenerate :
    (A : Set) -> (gs : List (Gen A)) -> (g : Gen A) -> (x : A) ->
    PropElem (Gen A) g gs ->
    CanGenerate' A g x    ->
    CanGenerate' A (oneof A gs) x
propOneofCanGenerate A gs g x el p =
  propBindCanGenerate (Gen A) A
    (elements (Gen A) gs) (\g' -> g')
    g                     x
    (propElementsCanGenerate (Gen A) gs g el)
    p
\end{code}

  For the proof that \<frequency\> preserves \<CanGenerate'\> we need
  some lemmas about the flattening of bags.

\begin{code}

propFlattenBagTail :
    (A : Set) -> (bag : Bag A) -> (x : A) -> (p : {Nat;A}) ->
    PropElem A x (flattenBag A bag) ->
    PropElem A x (flattenBag A (p::bag))
propFlattenBagTail A bag x (0,y) p        = p
propFlattenBagTail A bag x (succ n, y) p  =
  fol-plugin(propFlattenBagTail A bag x (n, y), p)

propFlattenBagElem :
    (A : Set) -> (bag : Bag A) ->
    (n : Nat) -> Not (Eq Nat n 0) ->
    (x : A) -> PropElem {Nat;A} (n,x) bag ->
    PropElem A x (flattenBag A bag)
propFlattenBagElem A [] n nz x p                = fol-plugin(p)
propFlattenBagElem A ((0,y)::bag) n nz x p      =
    fol-plugin( propFlattenBagElem A bag n nz x
              , p, nz, Nat
              , __fol_pair_injectivity
              )
propFlattenBagElem A ((succ m, y)::bag) n nz x p =
    fol-plugin( propFlattenBagElem A bag n nz x
              , propFlattenBagTail A bag x (m,y)
              , p
              , __fol_pair_injectivity
              )

\end{code}

  Now we're ready to prove \<CanGenerate'\> for \<frequency\>. In the case
  where \<frequency\> is applied to a concrete list this property isn't
  necessary, since its just as easy to work with \<oneof\> applied to the
  flattened list.

\begin{code}

propFrequencyCanGenerate :
    (A : Set) -> (fs : Bag (Gen A)) ->
    (n : Nat) -> Not (Eq Nat n 0) -> (g : Gen A) -> (x : A) ->
    PropElem {Nat;Gen A} (n,g) fs ->
    CanGenerate' A g x -> CanGenerate' A (frequency A fs) x
propFrequencyCanGenerate A fs n nz g x el p =
  propOneofCanGenerate A gs g x
    (propFlattenBagElem (Gen A) fs n nz g el)
    p
  where
    gs = flattenBag (Gen A) fs

\end{code}

  The most interesting combinator is of course \<sized\>. The property below
  isn't completely satisfactory, mainly because the requirement on \<f\> is not
  very nice (and quite strong?). It should be possible to come up with
  something better that still gives \<CanGenerate'\> for \<sized f\>.

\begin{code}

propSizedCanGenerate :
    (A : Set) -> (f : Nat -> Gen A) -> (x : A) ->
    (n : Nat) ->
    ( (n' : Nat) -> Holds (ordNat.leq n n') ->
      (size : Nat) -> CanProduce A (f n' size) x
    ) -> CanGenerate' A (sized' A f) x
propSizedCanGenerate A f x n p =
    (n, \n' leq -> p n' leq n')

propArbNatSCanProduce :
    (x : Nat) -> (n : Nat) -> Holds (ordNat.leq x n) ->
    (size : Nat) -> CanProduce Nat (arbNatS n size) x
propArbNatSCanProduce 0 0 leq size                = propReturnCanProduce Nat 0
propArbNatSCanProduce (succ x) 0 leq size         = (seed, fol-plugin(leq))
propArbNatSCanProduce 0 (succ n) leq size         =
    (propFrequencyCanGenerate Nat fs
        1 fol-plugin(Nat)
        ret 0
        fol-plugin( __fol_pair_injectivity
                  , __fol_extensionality
                  )
        (propReturnCanGenerate Nat 0)
    ).2 size tt
  where
    ret = return Gen monadGen Nat 0
    fs : Bag (Gen Nat)
    fs  = [ (1, ret)
          , (5, liftM Gen monadGen Nat Nat succ (arbNatS n))
          ]
propArbNatSCanProduce (succ x) (succ n) leq size  =
    (propFrequencyCanGenerate Nat fs
        5 fol-plugin(Nat)
        gensucc (succ x)
        fol-plugin( __fol_pair_injectivity
                  , __fol_extensionality
                  )
        (propLiftMCanGenerate Nat Nat succ (arbNatS n) x
          (size, \size' _ -> propArbNatSCanProduce x n leq size')
        )
    ).2 size  (propIf Nat (\z -> Holds (ordNat.leq z size))
                  (ordNat.leq size 0) 0 size
                  (\_ -> tt)
                  (\_ -> reflOrd Nat ordNat size)
              )
  where
    gensucc = liftM Gen monadGen Nat Nat succ (arbNatS n)
    fs : Bag (Gen Nat)
    fs  = [ (1, return Gen monadGen Nat 0)
          , (5, gensucc)
          ]

propArbNatCanGenerate : (n : Nat) -> CanGenerate' Nat arbNat n
propArbNatCanGenerate n =
  propSizedCanGenerate Nat arbNatS n n (propArbNatSCanProduce n)

{-
-- arbNatS is surjective
propSurjArbNatS : (n : Nat) -> (size : Nat) ->
                    CanProduce Nat (arbNatS n size) n
propSurjArbNatS 0 size        = (seed, reflexivity Nat 0)
propSurjArbNatS (succ n) size =
  (propSurjOneof Nat gs succGen (succ n) fol-plugin()
                (size, r', fol-plugin(ih.2, splitR.2))
  ).2
  where
    ih  = propSurjArbNatS n size

    succGen = liftM Gen monadGen Nat Nat succ (arbNatS n)
    bag : List {Nat; Gen Nat}
    bag   = [ (1, monadGen.return Nat 0)
            , (5, succGen)
            ]
    gs    = flattenBag (Gen Nat) bag

    splitR  = propSplitSurjective ih.1 seed
    r'      = splitR.1

-- arbNat is surjective
propSurjArbNat : (n : Nat) -> CanGenerate Nat arbNat n
propSurjArbNat n = (n, propSurjArbNatS n n)-}
\end{code}

