{-# OPTIONS -fglasgow-exts #-}

module Syntax.Internal
	( module Syntax.Name
	, module Syntax.Internal
	)
    where

import Data.Generics

import Utils.Map    ( Map )
import Utils.Pairs
import Utils.Generics

import Syntax.Name
import Syntax.Position

{--------------------------------------------------------------------------
    Terms
 --------------------------------------------------------------------------}

data AnnotatedTerm t v d
	= App (AnnotatedTerm t v d) (AnnotatedTerm t v d)
	| Var Name t	-- ^ name bound by lambda or let
	| Def Name d	-- ^ defined names
	| Lam Name (AnnotatedTerm t v d) t
	| Star Universe
	| Prop
	| Pi Name (AnnotatedTerm t v d) (AnnotatedTerm t v d) t
	| CallPlugin Name [Term] v
		-- ^ We don't require the arguments to a plugin to
		--   be type checked.
	| Let [AnnotatedLocalDefinition t v d]
	      (AnnotatedTerm t v d)
    deriving (Typeable, Data)

type Term	= AnnotatedTerm () () ()
type TypedTerm	= AnnotatedTerm Type Value DefValue

{--------------------------------------------------------------------------
    Local definitions
 --------------------------------------------------------------------------}

data AnnotatedLocalDefinition t v d
	= LocalDef  Name
		    (Maybe (AnnotatedTerm t v d))
		    (AnnotatedTerm t v d)
		    v
    deriving (Typeable, Data)

type LocalDefinition	    = AnnotatedLocalDefinition () () ()
type TypedLocalDefinition   = AnnotatedLocalDefinition Type Value DefValue


{--------------------------------------------------------------------------
    Universes
 --------------------------------------------------------------------------}

type Universe = Int

{--------------------------------------------------------------------------
    Definitions
 --------------------------------------------------------------------------}

type Program = [Definition]

data AnnotatedDefinition t v d
	= NormalDef Name (Maybe (AnnotatedTerm t v d))
			 (AnnotatedTerm t v d)
        | NewAxiom Name (AnnotatedTerm t v d)
    deriving (Typeable, Data)

type Definition		= AnnotatedDefinition () () ()
type TypedDefinition	= AnnotatedDefinition Type Value DefValue

data DefValue = Definition Value | Axiom Type
    deriving (Typeable, Data)

{--------------------------------------------------------------------------
    Views on terms
 --------------------------------------------------------------------------}

-- | Invariant: The 'Term' is not an application.
data TermAppView = TAVApp Term [Term]

tAppView :: Term -> TermAppView
tAppView t =
    case t of
	App t t'    ->
	    case tAppView t of
		TAVApp t0 ts	-> TAVApp t0 (ts ++ [t'])
	_	    -> TAVApp t []


{--------------------------------------------------------------------------
    Syntactic equality on terms
 --------------------------------------------------------------------------}

-- | This is the literal equality, ignoring position information
--   and annotations.
instance Eq (AnnotatedTerm t v d) where
    App a b		== App c d		= (a,b)		== (c,d)
    Var x _		== Var y _		= x		== y
    Def x _		== Def y _		= x		== y
    Lam x a _		== Lam y b _		= (x,a)		== (y,b)
    Star u		== Star v		= u		== v
    Prop		== Prop			= True
    Pi p a b _		== Pi q c d _		= (p,a,b)	== (q,c,d)
    CallPlugin x ts _	== CallPlugin y us _	= (x,ts)	== (y,us)
    Let defs a		== Let defs' b		= (defs,a)	== (defs',b)
    _			== _			= False

instance Eq (AnnotatedLocalDefinition t v d) where
    LocalDef x t e _	== LocalDef y t' e' _	= (x,t,e) == (y,t',e')

{--------------------------------------------------------------------------
    Values
 --------------------------------------------------------------------------}

data Value  = VId Name Type		-- ^ . @x :: V@
	    | VDef Name DefValue	-- ^ . @f^n :: V {clvs}@ (@n@ = arity, @clvs@ = clause values)
	    | VApp Value Value		-- ^ . @v w@ where @v@ is not a @VLam@
	    | VPi Value Closure Type	-- ^ . @(Fun W F) :: V@, type value @V@ must be a sort
	    | VLam Closure Type		-- ^ . @(Lam F) :: V@, type value @V@ must be a @Fun@
	    | VSort SortValue		-- ^ . user sort or @#n@.
	    | VTop Type			-- ^ . @Top :: V@
	    | VTag Value		-- ^ . Used internally when substituting inside closures.
					--     Should never appear outside the substitution function.
    deriving (Typeable, Data)

data SortValue	= VSUniv Universe
		| VSProp
    deriving (Typeable, Data)

type Type   = Value

data Closure = Cl { clName :: Name, clApp :: Value -> Value }
    deriving (Typeable, Data)

-- * Views on values 
----------------------------------------------------------------------

data ValueAppView
	= VAVId Name Type [Value]
	| VAVDef Name DefValue [Value]
	| VAVTop Type
	| VAVSort SortValue
	| VAVLam Closure Type
	| VAVPi Value Closure Type

vAppView :: Value -> ValueAppView
vAppView v =
    case v of
	VId x t		    -> VAVId x t []
	VDef x d	    -> VAVDef x d []
	VPi v cl t	    -> VAVPi v cl t
	VLam cl t	    -> VAVLam cl t
	VSort u		    -> VAVSort u
	VTop t		    -> VAVTop t
	VApp v w	    -> vavApp (vAppView v) w

vavApp :: ValueAppView -> Value -> ValueAppView
vavApp v w =
    case v of
	VAVId x t vs	-> VAVId x t (vs ++ [w])
	VAVDef x d vs	-> VAVDef x d (vs ++ [w])
	_		-> error "Internal error in vavApp!"

