{-| Bidirectional type checking and checking of a 
    program (= list of definitions).
-}

module TypeChecker where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad		( when, unless )
import Control.Monad.Error
import Data.Maybe               ( isJust, fromJust )
import qualified Data.Set as Set

import Plugins			( pluginCheck, pluginInfer )

import Syntax.Internal
import Syntax.Values
import Syntax.Position
import Syntax.Sorts
import Syntax.Pretty

import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Datatypes
import TypeChecker.Monad.Signatures
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Context
import TypeChecker.Monad.Options
import TypeChecker.Monad.MetaVars
import TypeChecker.Monad.Rewriting
import TypeChecker.Rewriting
import TypeChecker.Equality
import TypeChecker.Patterns
import TypeChecker.Patterns.Inductive
import TypeChecker.With
import TypeChecker.Definitions
import TypeChecker.Values
import TypeChecker.Values.Refresh
import TypeChecker.UnfoldConstants
import TypeChecker.Universes
import TypeChecker.Context
import TypeChecker.Errors
import TypeChecker.Abstractions
import TypeChecker.Sorts hiding (instantiate) --, (=:))
import TypeChecker.MetaVars
import TypeChecker.Values.SanityCheck

import TypeChecker.Sized -- for type-based termination

import Utils.Map as Map hiding (map)
import Utils.Names		( fresh )
import Utils.Monads
import Utils.Pairs
import Utils.Continuations

import Utils.Debug

-- Check program ----------------------------------------------------------

checkProgram :: Program -> TC ()
checkProgram [] = return ()
checkProgram (def : prog) =
    do	clearMetaVariables
	case def of
	    NormalDef r name h ar t cs	       -> checkDef r name h ar t cs
	    Synonym r name t		       -> checkSynonym r name t
	    NewAxiom  r name h t	       -> checkDef r name h h t []
	    NewData sized r name htel tel t cs -> checkData sized r name htel tel t cs
	    NewSig r name htel tel t ps cs     -> checkSig r name htel tel t ps cs
	    Rewrite r es		       -> mapM_ (checkRewrite r) es
	    Mutual r ds			       -> checkMutual r ds
	    AbstractDefs r ds		       -> checkAbstract r ds
	checkProgram prog

checkAbstract :: Range -> [Definition] -> TC ()
checkAbstract r ds =
    sayWhere r $
    sayWhen whenCheckingMutual $
    do	mapM_ checkDefType ds
	checkThatAllMetasHaveBeenSolved
	let xs = concatMap getDefName ds
	dvs <- mapM lookupDef xs
	mapM_ delDefinition xs
	checkProgram ds
	zipWithM_ removeDefinitionBody xs dvs
    where
	getDefName d = case d of
	    NormalDef _ x _ _ _ _ -> [x]
	    Synonym _ x _	  -> [x]
	    NewAxiom _ x _ _	  -> [x]
	    NewData _ _ x _ _ _ _ -> [x]
	    NewSig _ x _ _ _ _ _  -> [x]
	    Rewrite _ _		  -> []
	    Mutual _ ds		  -> concatMap getDefName ds
	    AbstractDefs _ ds	  -> concatMap getDefName ds

	removeDefinitionBody x dv =
	    do	addDefinition x dv
		delDatatype x
		delSignature x

checkMutual :: Range -> [Definition] -> TC ()
checkMutual r ds =
    sayWhere r $
    sayWhen whenCheckingMutual $
    do	mapM_ checkDefType ds
	mapM_ checkDefValue ds
	checkThatAllMetasHaveBeenSolved

checkDefType (NormalDef r name h ar t _) =
    sayWhere r $
    sayWhen (whenCheckingDefinition name) $
	do  defs <- getDefinitions
	    t' <- isType t
	    n <- nofBoundVars <$> getContext
	    addDefinition name (DefVal h ar n [] [] t') -- FREEVARS

checkDefType (NewData sized r name htel tel e _) =
    checkDataType r name htel tel e

checkDefType (NewSig r name htel tel e _ _) =
    checkSigType r name htel tel e

checkDefType _ = error "impossible: TypeChecker.checkDefType"

checkDefValue (NormalDef _ name _ _ _ cs) =
    do	DefVal h ar _ _ _ t <- lookupDef name
	checkClauses name h ar t [] cs

checkDefValue (NewData _ r name htel tel _ cs) =
    checkDataValue r name htel tel cs

checkDefValue (NewSig r name htel tel _ ps cs) =
    checkSigValue r name htel tel ps cs

checkDefValue _ = error "impossible: TypeChecker.checkDefValue"

-- Signatures
checkSigType :: Range -> Name -> Telescope -> Telescope -> Term -> TC ()
checkSigType r name htel tel e =
    sayWhere r $
    sayWhen (whenCheckingDefinition name) $
	do  let h  = length htel
		ar = h + length tel
	    t <- isType (telePi (htel ++ tel) e)

	    -- Check that the target is a sort
	    do	(n, t') <- endsIn t
		case vView $ unEl t' of
		    VSort s | n == ar
			    -> return ()
		    _	    -> failure $ notASortError e

	    n <- nofBoundVars <$> getContext
	    addDefinition name $ DefVal h ar n [] [] t -- FREEVARS

checkSigValue :: Range -> Name -> Telescope -> Telescope -> [Name] -> [Constructor] -> TC ()
checkSigValue r name htel tel ps cs =
    sayWhere r $
    sayWhen (whenCheckingDefinition name) $

    do	defval@(DefVal h ar _ _ _ t) <- lookupDef name

	-- The telescope is in scope when checking the constructors.
	VSort s <- vView . unEl . snd <$> endsIn t

	-- Check that the super signatures all have the same type
	do  ts <- map typeOf <$> mapM lookupDef ps
	    let check (p, t') = sayWhen (whenCheckingInheritance name p)
			      $ sayWhat subSignatureMustHaveSameType
			      $ t' =<= t    -- TODO: is this safe?
	    mapM_ check $ zip ps ts

	parents <- zip ps <$> mapM lookupSignature_ ps

	cs' <- insidePi (map fst $ htel ++ tel) t $ \ar w ->
	       computeInheritance 0 parents $ \inh ->
		do  cs' <- thread (checkField name s (htel ++ tel)) (zip [0..] cs) return
		    return $ inh ++ cs'

	addSignature name ps cs'

    where

	typeOf (DefVal _ _ _ _ _ t) = t

	computeInheritance :: Int -> [(Name, [(Name, DefValue)])]	->
			      ([(Name, FieldSource, DefValue)] -> TC a) ->
			      TC a
	computeInheritance _ [] ret = ret []
	computeInheritance n ((p,fs):ps) ret =
		computeInh n p (raiseBy n fs) $ \ (m, fs') ->
		computeInheritance m ps $ \ fss ->
		ret $ fs' ++ fss
	    where
		computeInh n p [] ret = ret (n, [])
		computeInh n p ((f,d):fs) ret =
		    do	mv <- lookupBoundVar f
			case mv of
			    Nothing -> extendContext f (Abstract $ typeOf d)
				       $ computeInh (n + 1) p fs
				       $ \ (n,fs') -> ret $ (n, (f, InheritedFrom p, d) : fs')
			    Just v  -> computeInh n p fsv ret
				where
				    TeleList fsv = instantiate (TeleList fs) [v]

checkSig :: Range -> Name -> Telescope -> Telescope -> Term -> [Name] -> [Constructor] -> TC ()
checkSig r name htel tel e ps cs =
    do	checkSigType r name htel tel e
	checkThatAllMetasHaveBeenSolved
	checkSigValue r name htel tel ps cs
	checkThatAllMetasHaveBeenSolved

checkField :: Name -> SortValue -> Telescope ->
	      (Int, Constructor) ->
	      ((Name, FieldSource, DefValue) -> TC a) -> TC a
checkField name s args (i, Constructor r conName tel e) ret =
    sayWhere r $
    sayWhen (whenCheckingConstructor conName) $

    do	-- Check that the type of the constructor fits in the
	-- type of the datatype.
	let e' = telePi tel e
	t <- isType e'
	typeType t =<= el (uSucc s) (vSort s)

	gamma <- reverse <$> getBoundContext
	let h = length args
	    n = length gamma

	(ar, _) <- endsIn t

	cv <- refresh t

	checkThatAllMetasHaveBeenSolved

	let defval = DefVal (length tel) ar n [] [] cv -- FREEVARS

	extendContext conName (Abstract t) $ ret (conName, NewField, defval)

checkDataType :: Range -> Name -> Telescope -> Telescope -> Term -> TC ()
checkDataType r name htel tel e =
    sayWhere r $
    sayWhen (whenCheckingDefinition name) $
	do  let h  = length htel
		ar = h + length tel
	    t <- isType (telePi (htel ++ tel) e)
	    n <- nofBoundVars <$> getContext
	    addDefinition name $ DefVal h ar n [] [] t -- FREEVARS

checkDataValue :: Range -> Name -> Telescope -> Telescope -> [Constructor] -> TC ()
checkDataValue r name htel tel cs =
    do	defval@(DefVal h ar _ _ _ t) <- lookupDef name
	-- The telescope is in scope when checking the constructors.
	cs' <- insidePi (map fst $ htel ++ tel) t $ \ar w ->
		    mapM (checkConstr name w (htel ++ tel) defval) cs

	-- Associate the constructors with the datatype.
	-- (checkConstr adds them as constants)
	addDatatype name cs'

checkData :: IsSized -> Range -> Name -> Telescope -> Telescope -> Term -> [Constructor] -> TC ()
checkData sized r name htel tel e cs =
    do	checkDataType r name htel tel e
	checkThatAllMetasHaveBeenSolved
	checkDataValue r name htel tel cs
	checkThatAllMetasHaveBeenSolved

checkRewrite :: Range -> Term -> TC ()
checkRewrite r e =
    sayWhere r $
    sayWhen (fsep $ pwords $ "checking that " ++ show e ++ " is a proof of a valid rewrite rule") $
    do	(_,t) <- inferType e
	rule  <- createRewriteRuleFromType t
	verbose 10 $ liftIO $ putStrLn $ "Adding global rewrite: " ++ show rule
	addGlobalRewriteRule rule

localRewrite :: Range -> Term -> TC a -> TC a
localRewrite r e k =
    sayWhere r $
    sayWhen (fsep $ pwords $ "checking that " ++ show e ++ " is a proof of a valid rewrite rule") $
    do	(_,t) <- inferType e
	rule  <- createRewriteRuleFromType t
	verbose 10 $ liftIO $ putStrLn $ "Adding local rewrite: " ++ show rule
	addLocalRewriteRule rule k

-- Go inside a Pi (brings abstracted variables in scope)
insidePi :: [Name] -> Type -> (Int -> Type -> TC a) -> TC a
insidePi [] t k		= k 0 t
insidePi (x:xs) t k =
    extendContext x (Abstract v) $
	insidePi xs (absBody f) $ \n -> k (n + 1)
    where
	VPi v f = vView (unEl t)


{- checkConstr: for example in

   data List(A:Set) : Set where
     nil  : List A
     cons : A -> List A -> List A

   Context = [ ("A", 'Set') ]
   w       = 'Set'
   dv      = DefVal 1 [] '(A::Set) -> Set'
   conName = "cons"
   t       = A -> List A -> List A
-}
checkConstr :: Name -> Type -> Telescope -> DefValue -> Constructor ->
	       TC (Name, DefValue)
checkConstr name w args dv@(DefVal _ dar _ _ _ _) (Constructor r conName tel e) =
    sayWhere r $
    sayWhen (whenCheckingConstructor conName) $

    do	-- Check that the type of the constructor fits in the
	-- type of the datatype.
	let e' = telePi tel e
	t <- isType e'
	(_, w') <- endsIn w	-- for inductive families
	typeType t =<= w'

	let VSort s = vView (unEl w')

	-- Check that the constructor returns an element
	-- of the datatype.
	(ar, t') <- endsIn t

	-- We check that we end in the datatype applied to something
	t'' <- unfoldConstants $ unEl t'
	case vAppView t'' of
	    VAVDef name' _ _
		| name == name' -> return ()
	    _			-> sayWhat (notConstructorTypeOf e w')
				   $ failure $ customError
				   $ "it doesn't return something in " ++ show name

	gamma <- reverse <$> getBoundContext
	let h  = length args
	    cv = piAbstraction (map snd $ take h gamma) t

	cv <- refresh cv

	checkThatAllMetasHaveBeenSolved

	n <- nofBoundVars <$> getContext
	let defval = DefVal (length tel + h) (h + ar) (n - h) [] [] cv -- FREEVARS

	addDefinition conName defval

	return (conName, defval)

checkTelescope :: Telescope -> (Int -> TelescopeValue -> TC a) -> TC a
checkTelescope [] k	     = k 0 []
checkTelescope ((x,e):tel) k =
    do	t <- isType e
	extendContext x (Abstract t) $
	    checkTelescope tel $ \n ts ->
		k (n + 1) (raiseBy n t : ts)

checkSynonym :: Range -> Name -> Term -> TC ()
checkSynonym r x e =
    sayWhere r $
    sayWhen (whenCheckingDefinition x) $
    do	defs <- getDefinitions
-- 	when (member x defs) $
-- 	    failure $ repeatedVariableError x
	(v,t) <- inferType e
	n <- nofBoundVars <$> getContext
	checkThatAllMetasHaveBeenSolved
	addDefinition x (DefVal 0 0 n [] [ClauseVal [] v] t)	-- FREEVARS

checkDef :: Range -> Name -> Hidden -> Arity -> Term -> [Clause] -> TC ()
checkDef r name h ar t cs =
    sayWhere r $
    sayWhen (whenCheckingDefinition name) $
	do  defs <- getDefinitions
-- We don't allow shadowed definitions.
-- This would lead to a problem in valueOf since definitions
-- don't carry closures.
-- TODO: They do now. Could we allow shadowing then?
-- We should never allow shadowed top level definitions. That's just plain stupid.
-- This should be checked when checking scopes though.
-- 	    when (member name defs) $
-- 		failure $ repeatedVariableError name

	    -- Check that the type is well formed
	    t' <- isType t
	    -- Check the clauses
	    checkClauses name h ar t' [] cs

checkClauses :: Name -> Int -> Arity -> Type -> [Value] -> [Clause] -> TC ()
checkClauses name h ar t scrutinees cs =

    do	n <- nofBoundVars <$> getContext
	addDefinition name (DefVal h ar n [] [] t)  -- FREEVARS

	-- Evaluated clauses have only type of recursive call
	vcs <- mapM (checkClause name t scrutinees) cs

	(t, vcs) <- refresh (t, vcs)

	unlessM isLocalDefinition checkThatAllMetasHaveBeenSolved

	-- Finally we can add the definition to the environment
	addDefinition name $ DefVal h ar n [] vcs t

inferClause :: Clause -> TC (ClauseValue, Type)
inferClause (Clause r _ [] (NoWith e)) =
    do	(v, t) <- inferType e
	return (ClauseVal [] v, t)
inferClause c@(Clause r _ _ _) =
    sayWhere r $ failure $ noInferenceForPatternClausesError c

checkClause :: Name -> Type -> [Value] -> Clause -> TC ClauseValue
checkClause name v scrutinees cl@(Clause r _ es0 with) =
	sayWhere r $
	    do	-- Expand constructor applications revealing hidden arguments
		verbose 6 $ liftIO $ putStrLn $ replicate 75 '-'
		verbose 6 $ liftIO $ putStrLn $ "checking patterns " ++ show es0
		es <- noWildCards =<< expandHidden es0
		verbose 6 $ liftIO $ putStrLn $ "expanded: " ++ show es
		-- Replace all variables (and non-patterns) by metas.
		es' <- variablesAsMetas es
		verbose 8 $ liftIO $ putStrLn $ "metified: " ++ show es'
		-- Typecheck the resulting patterns.
		(vts,_) <- allowUnsolvedMetas $ fits es' v
		verbose 6 $ liftIO $ putStrLn $ "values  : " ++ show vts
		-- Compute the relation between named variables and meta variables.
		-- Checks that a variable doesn't have conflicting values.
		(es1, metaMap) <- matchPattern es vts
		verbose 6 $ liftIO $ putStrLn $ "updated : " ++ show es1
		verbose 6 $ liftIO $ putStrLn $ "metamap : " ++ show metaMap
		-- Compute the bound variables. Instantiates the meta variables with
		-- the corresponding variables. Also returns the resulting patterns.
		(tel, es2, ps) <- computeBindings metaMap es1 (map fst vts)
		verbose 6 $ liftIO $ putStrLn $ "bindings: " ++ show tel
		verbose 6 $ liftIO $ putStrLn $ "patterns: " ++ show ps
		-- Type check the original pattern
		case with of
		    NoWith e	->
			extendContextWithTel tel $
			    do	--let es2 = dropUnderscores es1 ps
				let n  = length tel
				    v' = raiseBy n v
				verbose 10 $ liftIO $ putStrLn $ "checking that " ++ show es2 ++ " fits " ++ show v'
				verbose 10 $
				    do  gamma <- getContext
					liftIO $ putStrLn $ "Context: " ++ show gamma
				(wts',w) <- fits es2 v'
				let ws		= map fst wts'
				    scrutinees'	=
					instantiateUnder n (raiseBy n scrutinees)
					$ drop (length scrutinees)
					$ reverse ws

				verbose 10 $ liftIO $ putStrLn $ "n = " ++ show n
				verbose 10 $ liftIO $ putStrLn $ "ws = " ++ show ws
				verbose 10 $ liftIO $ putStrLn $ "raised = " ++ show (raiseBy n scrutinees)
				verbose 8 $ liftIO $ putStrLn $ "scrutinees: " ++ show scrutinees
				verbose 8 $ liftIO $ putStrLn $ "scrutinees': " ++ show scrutinees'
				verbose 10 $ liftIO $ putStrLn $ "it does: remainder = " ++ show w

				w' <- refresh w
				verbose 10 $ liftIO $ putStrLn $ "after refreshing: " ++ show w'
				u <- rewriting scrutinees'
					(drop (length ws - length scrutinees') ws)
					$ lamAbstraction n <$> checkType e w'
				--u' <- refresh u
				verbose 10 $ liftIO $ putStrLn $ "clause body: " ++ show u
				return $ ClauseVal ps u
			where
			    rewriting [] [] k = k
			    rewriting (v:vs) (w:ws) k =
				do  mrule <- (Just <$> createRewriteRule 0 v w)
						`catchError` \_ -> return Nothing
				    case mrule of
					Nothing	-> do	verbose 8 $ liftIO $ putStrLn $ show v ++ " --> " ++ show w ++ " is not a valid rewrite rule"
							rewriting vs ws k
					Just r	-> do	verbose 8 $ liftIO $ putStrLn $ "case rewrite rule: " ++ show r
							addLocalRewriteRule r $ rewriting vs ws k
			    rewriting _ _ _ = error "impossible: TypeChecker.rewriting"
		    With es cs	->
			do  (vs,t) <- extendContextWithTel tel $
					do  --let es2 = dropUnderscores es1 ps
					    (_,w) <- fits es2 v
					    (vs,ts) <- unzip <$> mapM inferType es
					    let t = createAuxType tel ts w
					    return (vs, t)
			    verbose 8 $ liftIO $ putStrLn $ "scrutinees: " ++ show vs
			    verbose 8 $ liftIO $ putStrLn $ "type of aux function: " ++ show t
			    aux	    <- fresh $ nameString name ++ "#"
			    cs'	    <- mapM (restrictClause ps) cs
			    verbose 8 $ liftIO $ putStrLn $ "restricted clauses: " ++ show cs'
			    clearLocalRewriteRules $
				checkClauses aux 0 (length tel + length es) t vs cs'
			    let	u = lamAbstraction (length tel)
						   (callAuxFunction aux tel vs)
			    return $ ClauseVal ps u


fits :: [Term] -> Type -> TC ([(Value, Type)], Type)
fits [] t = return ([], t)
fits (e:es) t =
    do	tt <- unfoldConstants (unEl t)
	case vView tt of
	    VPi w f ->
		do  u	    <- checkType e w
		    (us,t') <- fits es (absApp f u)
		    return ((u,w):us, t')
	    _	-> failure $ tooManyArgumentsError t

checkPatternSpine :: Type -> [Pattern] -> Type ->
		     (Int -> [Value] -> TC a) -> TC a
checkPatternSpine t [] w k =
    do	t =<= w
	k 0 []
checkPatternSpine t (p:ps) w k =
    do	tt <- unfoldConstants (unEl t)
	case vView tt of
	    VPi u f ->
		checkPattern p u $ \n pv ->
		do  let (f',w') = raiseBy n (f,w)
			b	= absApp f' pv
		    checkPatternSpine b ps w' $ \m pvs ->
			k (n + m) (raiseBy m pv : pvs)
	    _	-> failure $ notAFunctionError (Var noRange $ dummyName "dummy") t
					-- TODO: very bad msg

checkPattern :: Pattern -> Type -> (Int -> Value -> TC a) -> TC a
checkPattern (WildP _) v k =
    do	x <- fresh "_"
	extendContext x (Abstract v) (k 1 $ vBId 0)

checkPattern p@(VarP x) v k =
    sayWhat (checkingPatternFailed p v) $
	extendContext x (Abstract v) (k 1 $ vBId 0)

checkPattern p@(PairP _ p1 p2) t k =
    sayWhat (checkingPatternFailed p t) $
    do	tt <- unfoldConstants (unEl t)
	case vView tt of
	    VSigma w f    ->
		checkPattern p1 w $ \n v1 ->
		    checkPattern p2 (absApp (raiseBy n f) v1) $ \m v2 ->
			k (n + m) (vPair (raiseBy m v1) v2)
	    _ -> failure $ patternHasSigmaTypeError p t

checkPattern p@(ConP _ c [] ps) t k =
    sayWhat (checkingPatternFailed p t) $
    do	(d,vs,cs) <- getDatatype (unEl t)
	unless (c `elem` cs) $ failure $ becauseWrongConstructor c d
	cdef@(DefVal h ar _ _ [] tc) <- lookupDef c
	unless (ar - h == length ps) $ failure $ becauseWrongNumberOfConstructorArguments c
	w' <- piApps tc vs
	checkPatternSpine w' ps t $
	    \n ws -> k n $ vApps (vCon c h ar) (raiseBy n vs ++ ws)
checkPattern p t _ =
    sayWhat (checkingPatternFailed p t) $
	failure $ customError $ "Irrefutable patterns cannot reveal hidden arguments"


-- | TODO: Move this somewhere else (TypeChecker.Monad.Datatype)
getDatatype :: Value -> TC (Name, [Value], [Name])
getDatatype v =
    do	v' <- unfoldConstants v
	case vAppView v' of
	    VAVDef x _ vs ->
		do  cs <- lookupDatatype x
		    return (x,vs,map fst cs)
	    _	-> failure $ becauseNotADatatype v

checkHSpine :: Term -> [Term] -> Type -> TC [Value]
checkHSpine hd [] t = return []
checkHSpine hd (e:es) t =
    do	tt <- unfoldConstants (unEl t)
	case vView tt of
	    VPi w _ -> do   tv	<- checkType e w
			    tvs <- checkHSpine (HApp hd e) es
						=<< piApp t tv
			    return $ tv:tvs
	    _	    -> failure $ notAFunctionError hd t

-- Type inference ---------------------------------------------------------

inferPi :: Pattern -> Term -> Term -> TC (Value, Type)
inferPi p a b =
    do	t1 <- isType a
	checkPattern p t1 $ \n _ ->
	    do  t2 <- isType b
-- 		va' <- unfoldConstants $ unEl va
		let s = piSort (getSort t1) (getSort t2)

		-- Explanation: See checking for lambda.
		x <- fresh $ patName p
		let vs	    = patternInstantiation p
		    t2'	    = raiseByFrom 1 (length vs) t2
		    t2''    = instantiate t2' vs
		return ( vPi t1 (mkAbs (nameString x) t2'')
		       , el (uSucc s) (vSort s)
		       )
--     where
-- 	piSort' sa sb = setSortIndex sb $ siMax (getSortIndex sa)
-- 					        (getSortIndex sb)

inferSigma :: Name -> Term -> Term -> TC (Value, Type)
inferSigma x a b =
    do	t1 <- isType a
	extendContext x (Abstract t1) $
	    do  t2 <- isType b
		let s = sigmaSort (getSort t1) (getSort t2)
		return	( vSigma t1 (mkAbs (nameString x) t2) 
			, el (uSucc s) $ vSort s
			)

-- | The third argument is the list of hidden arguments given explicitly.
inferDef :: Range -> Name -> [Term] -> (Int -> Arity -> Value) -> TC (Value, Type)
inferDef r x es mkVal =
    do	d <- lookupDef x
	inferHiding r (Def r x) es d mkVal

inferHiding :: Range -> Term -> [Term] -> DefValue -> (Int -> Arity -> Value) -> TC (Value, Type)
inferHiding r e es d@(DefVal h ar fv _ _ t) mkVal =
    do	verbose 10 $ liftIO $ putStrLn $ "inferHiding " ++ show e ++ " : " ++ show d
	vs <- checkHSpine e es t
	let n = length es
	ws <- replicateM (h - n) createMeta
	zipWithM (\n w -> newMetaVariable r e n w) [n + 1..] ws
	nbound <- nofBoundVars <$> getContext
	let fvs = map vBId [nbound - 1, nbound - 2 .. nbound - fv]
	let v = mkVal (h + fv) ar `vApps` (fvs ++ vs ++ ws)
	t'' <- piApps t (vs ++ ws)
	verbose 12 $
	    case vAppView $ unEl t'' of
		VAVDef y _ vs'	->
		    liftIO $ putStrLn $ "inferHiding " ++ show e ++ " : " ++ show t ++ " @ " ++ show (vs ++ ws) ++ " --> " ++ show y ++ " " ++ show vs'
		_   -> return ()
	return (v,t'')

inferProj r e0 e x es =
    do	(v,t) <- inferType e
	t' <- unfoldConstants $ unEl t
	case vAppView t' of
	    VAVDef c _ vs   ->
		do  cs <- lookupSignature_ c
		    d  <- fieldType c v (reverse vs) x cs
		    inferHiding r e0 es d (\_ _ -> v `vProject` VField x)
	    _		    -> failure $ notAStructError e t'
    where
	fieldType c v vs x [] = failure $ noSuchFieldError x v c
	fieldType c v vs x ((y,d):ds)
	    | x == y'	=
		do  verbose 10 $ liftIO $ putStrLn $ "field " ++ x ++ ": " ++ show vs ++ " -> " ++ show d
		    let DefVal h ar _ is cs t = instantiate d vs
		    return $ DefVal h ar 0 is cs t
	    | otherwise	= fieldType c v (vProject v (VField y') : vs) x ds
	    where y' = nameString y

inferType :: Term -> TC (Value, Type)
inferType e =
	sayWhen (whenInferringType e) $
	sayWhere (getRange e) $
	    inferType' e
    where

	vDef' x h _ = vDef x h

	inferType' (Var _ x) = sayWhat doingNothingMuch $ lookupVar x
	inferType' (Con r x)	= inferDef r x [] (vCon x)
	inferType' (Def r x)	= inferDef r x [] (vDef' x)
	inferType' (Data r x)	= inferDef r x [] (vDef' x)
	inferType' (Sig r x)	= inferDef r x [] (vDef' x)

	inferType' e0@(Project r e (Field _ x)) = inferProj r e0 e x []

	inferType' e0@(HApp t' _) =
	    case tAppView e0 of
		TACon  r x   ts [] -> inferDef r x ts (vCon x)
		TADef  r x   ts [] -> inferDef r x ts (vDef' x)
		TAData r x   ts [] -> inferDef r x ts (vDef' x)
		TASig  r x   ts [] -> inferDef r x ts (vDef' x)
		TAProj r e x ts [] -> inferProj r e0 e x ts
		_		   -> failure $ hiddenApplicationError t'

	inferType' e@(App e1 e2) =
	    do	(v1,t1) <- inferType e1
		t1' <- unfoldConstants (unEl t1)
		case vAppView t1' of
		    VAVPi t2 f	-> checkArg v1 t2 f
		    VAVMeta m vs ->
			do  instantiateToPi m (getSort t1) vs
			    t1'' <- unfoldConstants t1'
			    case vAppView t1'' of
				VAVPi t2 f  -> checkArg v1 t2 f
				_	    -> failure $ customError $ "panic: instantiateToPi should give a Pi, not " ++ show t1''
		    _		-> failure $ notAFunctionError e1 t1
	    where
		checkArg v1 t2 f =
		    do v2 <- checkType e2 t2
		       return (vApply v1 v2, absApp f v2)

	inferType' e0@(Project _ e p) =
	    do	(v,t) <- inferType e
		t'    <- vAppView <$> unfoldConstants (unEl t)
		inferProj v t p t'
	    where
		inferProj v t (Proj r n) (VAVSigma w f) =
		    case n of
			n | n < 1   -> failure $ customError $ "panic: non-positive projection from " ++ show v
			1	    -> return (first, w)
			2	    -> return (second, absApp f first)
			n	    ->
			    do	let t = absApp f first
				t' <- vAppView <$> (unfoldConstants $ unEl t)
				inferProj second t (Proj r (n - 1)) t'
		    where
			first = vProject v VFst
			second = vProject v VSnd

		inferProj v t (Proj r n) (VAVMeta m vs) =
		    do	instantiateToSigma m (getSort t) vs
			t' <- vAppView <$> unfoldConstants (unEl t)
			inferProj v t (Proj r n) t'

		inferProj _ t (Proj _ _) _  = failure $ notAPairError e t

		inferProj v t (Field r x) _ =
		    sayWhere r $ failure $ customError $ "Only structs have named fields, and they are temporarily out of commission"

	inferType' (Hash _ u)	= return (vSort s, sortType s)
	    where
		s = VSUniv u
	inferType' (Star _ u)	=
	    do	s <- VSStar . plus u <$> freshSortIndex
		return (vSort s, sortType s)
	    where
		plus UZero i	    = i
		plus (USucc u) i    = siSucc (plus u i)

	inferType' (ISet _)	= return (vSort VSISet, sortType VSISet)
	inferType' (Prop _)	= return (vSort VSProp, sortType VSProp)

	inferType' (Pi _ x a b)	    = inferPi x a b
	inferType' (Sigma _ x a b)  = inferSigma x a b

	inferType' (Let r defs e) = typeLet defs $ inferType e

	{-  It is crucial for correctness that we don't allow inference for
	    local rewrites. If we do, rewriting doesn't preserve typeability.
	    Example:
		r		: A == A'
		f		: A -> B
		e		: A
		f e		: B
		rewrite r in f	: A' -> B
		which cannot be applied to e, since without eq A doesn't reduce to A'.
	-}
-- 	inferType' (RewriteIn r es e) =
-- 	    foldr (.) id (map (localRewrite r) es) $ inferType e

	inferType' (CallPlugin r scope name es) =
	    withScopeEnv scope $ pluginInfer name r es

	inferType' e = failure $ cannotInferError e

-- | Get the type of a sort
sortType :: SortValue -> Type
sortType s = el s2 (vSort s1)
    where
	s1 = uSucc s
	s2 = uSucc s1

-- | Get the type of a type
typeType :: Type -> Type
typeType t = el (uSucc s) (vSort s)
    where
	s = getSort t

-- Type checking ----------------------------------------------------------

-- | invariant: the 'Type' should be a type value and
-- the context should consist of (isType)s. If
-- @w <- checkType e v@ then @valueType w@ is a subtype of @v@.
checkType :: Term -> Type -> TC Value
checkType t v =
	sayWhen (whenCheckingType t v) $
	sayWhere (getRange t) $
	    checkType' t =<< typeMapM unfoldConstants v
    where

	checkType' :: Term -> Type -> TC Value
	checkType' ee@(Lam _ p e) v =
	    case vAppView $ unEl v of
		VAVPi w f ->
		    do	u <- checkPattern p w $ \n pv ->
				checkType e $ absApp (raiseBy n f) pv
			x <- fresh $ patName p

			{- Explanation:
			    u is valid in the context G, expand p
			    the body of the lambda (that we return) u'
			    should be valid in G, x. We have to instantiate
			    the parameters of u with the proper projections
			    from x.
			-}
			-- Intuitively we work inside the closure here
			let vs = patternInstantiation p
			    u' = raiseByFrom 1 (length vs) u
			    u0 = instantiate u' vs
			return $ vLam (mkAbs (nameString x) u0)

		VAVMeta m vs	->
		    do	instantiateToPi m (getSort v) vs
			checkType ee v
		_ -> failure $ lambdaHasFunctionTypeError ee v

	checkType' e@(Pair _ s t) v =
	    case vAppView $ unEl v of
		VAVSigma w f ->
		    do	vs <- checkType s w
			vt <- checkType t (absApp f vs)
			return $ vPair vs vt
		VAVMeta m vs	->
		    do  instantiateToSigma m (getSort v) vs
			checkType e v
		_ -> failure $ hasSigmaTypeError e v

	checkType' e@(Struct r xs) v =
	    case vAppView $ unEl v of
		VAVDef x _ vs	->
		    do	cs <- lookupSignature_ x
			bs <- checkStruct (reverse vs) xs cs
			return $ vStruct bs

		_   -> failure $ hasSignatureTypeError e v
	    where
		checkStruct vs ((x,hps,ps,e):xs) ((y,DefVal h _ _ _ _ t):cs)
		    | nameString x /= nameString y  = failure $ structLabelMismatchError x y
		    | length hps > h		    = failure $ tooManyHiddenArguments x (length hps) h
		    | otherwise		=
			do  pad <- map VarP <$> replicateM (h - length hps) (fresh "_")
			    let e' = buildLam (hps ++ pad ++ ps) e
				t' = instantiate t vs
			    v  <- checkType e' t'
			    bs <- extendContext x (Defined v t') $ checkStruct (v:vs) xs cs
			    return $ (nameString x, v) : bs

		checkStruct _ [] []		= return []
		checkStruct _ ((x,_,_,_):_) []	= failure $ tooManyFieldsInStructError x
		checkStruct _ [] cs		= failure $ tooFewFieldsInStructError $ map fst cs

	checkType' (CallPlugin r scope name es) v =
	    withScopeEnv scope $ pluginCheck name r es v

	-- A meta variable can have any type.
	checkType' (Meta r _) _ =
	    do	v <- createMeta
		newMetaVariable r (Var r DontCare) 0 v
		return v

	-- Lets can be both inferred and checked, depending on whether the
	-- body is inferrable or not.
	checkType' (Let _ defs e) v = typeLet defs $ checkType e v

	checkType' (RewriteIn r es e) v =
	    foldr (.) id (map (localRewrite r) es) $ checkType e v

	-- Anything else can be inferred.
	checkType' e v =
	    do	(ev,w) <- inferType e
		sayWhat (checkingSubtypingConstraint w v) $
		    w =<= v
		return ev

typeLet :: [Definition] -> TC a -> TC a
typeLet ds k = fancyLet ds k --normalLet ds k

-- | Allowing arbitrary definitions in local declarations.
fancyLet :: [Definition] -> TC a -> TC a
fancyLet ds k =
    do	localDefinition $ checkProgram ds
	k

-- | The normal let rule. The body of a local definition is known whenever
--   the definition is in scope.
normalLet :: [LocalDefinition] -> TC a -> TC a

normalLet [] k = k

normalLet (LocalDef _ x Nothing _ b : defs) k =
    do	(v,t) <- inferType b
	extendContext x (Defined v t) $
	    normalLet defs k

normalLet (LocalDef _ x (Just e) _ b : defs) k =
    do	t <- isType e
	w <- checkType b t
	extendContext x (Defined w t) $
	    normalLet defs k

-- Is type ----------------------------------------------------------------

isType :: Term -> TC Type
isType e =
    do	(v, t) <- inferType e
	t' <- unfoldConstants (unEl t)
	case vView t' of
	    VSort s -> return $ el s v
	    _	    -> failure $ notATypeError e t

isSort :: Term -> TC SortValue
isSort e =
    do	(v, _) <- inferType e
	v' <- unfoldConstants v
	case vView v' of
	    VSort s -> return s
	    _	    -> failure $ notASortError e

