{-# OPTIONS -fglasgow-exts #-}

module TypeChecker.Values.SanityCheck where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad.Reader
import Data.IORef

import Syntax.Internal
import Syntax.Values
import Syntax.Pretty

import TypeChecker.Values
import TypeChecker.Universes
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.Monad.Context
import TypeChecker.Monad.Options
import TypeChecker.Abstractions
import TypeChecker.UnfoldConstants
import TypeChecker.Errors
import TypeChecker.Context

import Utils.Error
import Utils.Monads

report :: Doc -> TC a -> TC a
report msg m =
    do	--liftIO $ print msg
	local (updateErrorMsg $ \e -> e { errWhen = errWhen e $$ msg }) m

insanity :: Doc -> TC ()
insanity msg =
    failure $ \e -> e { errWhat = fsep $ pwords "insanity detected"
		      , errWhy = msg
		      }

-- | Syntactic equality on values
-- instance Eq Value where
--     VId x t		== VId x' t'		    = (x,t)	    == (x',t')
--     VBId n t		== VBId n' t'		    = (n,t)	    == (n',t')
--     VMeta m t		== VMeta m' t'		    = (m,t)	    == (m',t')
--     VDef x d		== VDef x' d'		    = (x,d)	    == (x',d')
--     VPi v f t		== VPi v' f' t'		    = (v,f,t)	    == (v',f',t')
--     VLam f		== VLam f'		    = f		    == f'
--     VSort s		== VSort s'		    = s		    == s'
--     VTop t		== VTop t'		    = t		    == t'
--     VPair v w t		== VPair v' w' t'	    = (v,w,t)	    == (v',w',t')
--     VSigma x w y f t	== VSigma x' w' y' f' t'    = (x,w,y,f,t)   == (x',w',y',f',t')
--     VApp v w		== VApp v' w'		    = (v,w)	    == (v',w')
--     VProj v p		== VProj v' p'		    = (v,p)	    == (v',p')
--     v == w = case map vAppView [v,w] of
-- 		VAVMeta _ _ _ : _	-> True
-- 		_ : VAVMeta _ _ _ : _	-> True
-- 		_			-> False

-- instance Eq Closure where
--     Cl _ t b	== Cl _ t' b'	= (t,b) == (t',b')
    
instance Eq DefValue where
    DefVal h ar _ _ _ t == DefVal h' ar' _ _ _ t' = (h,ar,t) == (h',ar',t')

instance Eq ClauseValue where
    ClauseVal ps t == ClauseVal ps' t'	= (ps,t) == (ps',t')

sanityCheck :: a -> TC ()
sanityCheck _ = return ()

{-

-- | Check that a value is well formed in the current context.
--   Assumption: Everything in the signature is sane. Not necessarily
--		 true for the context.
class Sane a where
    sanityCheck :: a -> TC ()

instance Sane a => Sane [a] where
    sanityCheck = whenM doSanityChecks . mapM_ sanityCheck

instance Sane Closure where
    sanityCheck cl =
	whenM doSanityChecks $
	do  report (text "looking at the domain") $ sanityCheck v
	    extendContext x (Abstract v) $ sanityCheck w
	where
	    x = clName cl
	    v = clType cl
	    w = clBody cl

instance Sane DefValue where
    sanityCheck d@(DefVal h ar _ cs t) =
	whenM doSanityChecks $
	report (text "checking defvalue" <+> pretty d) $
	do  report (text "looking at type") $ sanityCheck t
	    report (text "looking at clauses") $ sanityCheck cs
	    unless (h <= ar)			$ insanity $ text "too many hidden arguments"
	    pis <- countPis t
	    unless (pis >= ar)			$ insanity $ text "too many arguments"
	    unless (all (ar==) $ map nPats cs)	$ insanity $ text "wrong number of patterns"
	where
	    countPis t =
		do  t' <- vView =<< unfoldConstants t
		    case t' of
			VPi w cl _  -> (1 +) <$> countPis (clBody cl)
			_	    -> return 0
	    nPats (ClauseVal ps _) = length ps

instance Sane ClauseValue where
    sanityCheck (ClauseVal _ v) = whenM doSanityChecks $ sanityCheck v

instance Sane Value where
    sanityCheck v = whenM doSanityChecks $ sane =<< unfoldConstants v
	where
	    sane v =
		report (text "checking" <+> pretty v) $
		vView v >>= \vv ->
		case vv of

		    VId x t	-> namedVariable x

		    VBId n t	->
			do  lookAtType $ sanityCheck t
			    w <- lookupBVar n
			    w' <- vView w
			    case w' of
				VBId m t'
				    | n /= m	-> insanity $ text "context mismatch"
				    | t /= t'	-> contextTypeMismatch t t'
				    | otherwise	-> return ()
				_   ->
				    report (text "checking definition") $
				    do  sanityCheck w
					t' <- valueType w
					unless (t == t') $ contextTypeMismatch t t'

		    VMeta m t	    ->
			do  lookAtType $ sanityCheck t
			    mv <- liftIO $ readIORef m
			    case mv of
				Open _		-> return ()
				Instantiated v	->
				    report (text "instantiated meta") $
				    do	sanityCheck v
					t' <- unfoldConstants t
					tv <- unfoldConstants =<< valueType v
					unless (t' == tv) $ metaVarTypeMismatch t' tv

		    VDef x d	    -> return ()

		    VPi v cl t ->
			do  lookAtType $ sanityCheck t
			    sanityCheck v
			    sanityCheck cl
			    unless (v == clType cl) $ closureTypeMismatch "the Pi" v (clType cl)
			    checkSorts v (clBody cl) t

		    VLam cl	    ->
			do  sanityCheck cl
			    lookAtType $ sanityCheck =<< valueType v

		    VSort s		    -> return ()
		    VTop t		    -> lookAtType $ sanityCheck t

		    VPair u v t	    ->
			do  lookAtType $ sanityCheck t
			    sanityCheck u
			    sanityCheck v
			    VSigma _ w _ cl _ <- vView =<< unfoldConstants t
			    ut <- unfoldConstants =<< valueType u
			    uv <- unfoldConstants =<< valueType v
			    w' <- unfoldConstants w
			    ct <- unfoldConstants =<< clApp cl u
			    let r   | ut /= w'  = pairTypeMismatch "first" ut w'
				    | uv /= ct  = pairTypeMismatch "second" uv ct
				    | otherwise = return ()
			    r

		    VSigma x w y cl t    ->
			do  lookAtType $ sanityCheck t
			    sanityCheck w
			    sanityCheck cl
			    unless (w == clType cl) $ closureTypeMismatch "the Sigma" w
									  (clType cl)
			    checkSorts v (clBody cl) t

		    VApp v w	->
			do  vv <- vView v
			    case vv of
				VTop _ -> insaneApp v
				VLam _ -> insaneApp v
				_      ->
				    do  sanityCheck v
					sanityCheck w
					t <- valueType (vApp v w)
					lookAtType $ sanityCheck t
					vt <- vView =<< unfoldConstants =<< valueType v
					case vt of
					    VPi u cl _  ->
						do  wt <- unfoldConstants =<< valueType w
						    t' <- unfoldConstants t
						    u' <- unfoldConstants u
						    ct <- unfoldConstants =<< clApp cl w
						    let r | u' /= wt  = appTypeMismatch "argument" wt u'
							  | t' /= ct  = appTypeMismatch "result" t' ct
							  | otherwise = return ()
						    r
					    _		-> shouldBePi (unView vt)
				

		    VProj v p	->
			do  vv <- vView v
			    case vv of
				VTop _      -> insaneProj v
				VPair _ _ _ -> insaneProj v
				_	    ->
				    do  sanityCheck v
					t <- vView =<< unfoldConstants =<< valueType v
					case t of
					    VSigma _ _ _ _ _ ->
						lookAtType $ sanityCheck
							   =<< valueType (vProj v p)
					    _ -> shouldBeSigma $ unView t

		where

		    namedVariable x =
			insanity $
			    fsep $ pwords "there is a named variable" ++ [text x]

		    contextTypeMismatch t t' =
			insanity $
			    fsep $ pwords "the type" ++ [pretty t] ++
				   pwords "doesn't match the type in the context" ++
				   [pretty t']

		    metaNotRefreshed = insanity $ fsep $ pwords "the meta variable was not refreshed"

		    definitionMismatch d d' =
			insanity $ fsep $ pwords "the definition" ++ [pretty d] ++
					  pwords "does not match the one in scope" ++
					  [pretty d']

		    closureTypeMismatch s v v' =
			insanity $ fsep $ pwords s ++ pwords "says the domain is" ++ [pretty v] ++
					  pwords "but the closure says" ++ [pretty v']

		    sortMismatch v v' =
			insanity $ fsep $ pwords "the sort was" ++ [pretty v] ++
					  pwords "but should have been" ++ [pretty v']

		    shouldBeSort msg v =
			insanity $ fsep $ pwords msg ++ [pretty v] ++ pwords "should be a sort"

		    pairTypeMismatch p t t' =
			insanity $ fsep $ pwords "the type of the" ++ [text p] ++
					  pwords "component is" ++ [pretty t] ++
					  pwords "but the type annotation says" ++
					  [pretty t']

		    metaVarTypeMismatch t t' =
			insanity $ fsep $ pwords "the type of the meta variable" ++
					  [pretty t] ++
					  pwords "is not the same as the type of its instantiation" ++
					  [pretty t']

		    insaneApp v = 
			insanity $ fsep $ [pretty v] ++ pwords "cannot be the left argument of an application"

		    insaneProj v = 
			insanity $ fsep $ [pretty v] ++ pwords "cannot be in a projection"

		    shouldBePi t =
			insanity $ fsep $ pwords "the type of the function is" ++ [pretty t] ++
					  pwords "which is not a function type"

		    shouldBeSigma t =
			insanity $ fsep $ pwords "the type of the projectee" ++ [pretty t] ++
					  pwords "should have been a sigma type"

		    appTypeMismatch s t t' =
			insanity $ fsep $ pwords "the" ++ pwords s ++ pwords "type" ++
					  [pretty t] ++ pwords "was claimed to be" ++
					  [pretty t']
					

		    lookAtType	    = report (text "looking at the type")

		    checkSorts v w t =
			    report (text "checking the sorts") $
				do  ts <- mapM (vView <.> unfoldConstants <.> valueType) [v,w]
				    case ts of
					[ VSort s1, VSort s2 ]
					    | t /= t'	-> sortMismatch t t'
					    | otherwise	-> return ()
					    where
						t' = vSort (piSort s1 s2)
					[ VSort _, _ ]	-> shouldBeSort "the type of the body" (unView $ ts!!1)
					[ _, _ ]	-> shouldBeSort "the type of the domain" (unView $ ts!!0)

-}
